#include <utlbuffer.h>
#include <windows.h>

#include <GMLuaModule.h>
#include "gm_aigraph.h"

#define	 AINET_VERSION_NUMBER	37
#define	 NUM_HULLS				10
#define  MAX_NODES				1500

// lua_run require("aigraph") nodes, links = ParseGraphFile("garrysmod/maps/graphs/gm_construct.ain")

int ParseGraph(CUtlBuffer &buf, ILuaInterface *gLua)
{
	int ainet_ver = buf.GetInt();
	int map_ver = buf.GetInt();

	if ( ainet_ver != AINET_VERSION_NUMBER)
	{
		gLua->Msg("Unknown graph file.\n");
		return 0;
	}

	int numNodes = buf.GetInt();
	if ( numNodes > MAX_NODES || numNodes < 0 )
	{
		gLua->Msg("Graph file has an unexpected amount of nodes\n");
		return 0;
	}

	// we get a reference to the function to survive past 510 calls!
	ILuaObject *vectormeta = gLua->GetGlobal("Vector");
	vectormeta->Push();
	int vectormetaref = gLua->GetReference(-1, true);
	vectormeta->UnReference();

	// we push the table on the stack to survive past 510 calls!
	ILuaObject *table = gLua->GetNewTable();
	table->Push();
	table->UnReference();

	for(int i=0; i < numNodes; i++)
	{
		ILuaObject *node = gLua->GetNewTable();

		Vector vector(buf.GetFloat(), buf.GetFloat(), buf.GetFloat());

		float yaw = buf.GetFloat();
		float flOffsets[NUM_HULLS];
		buf.Get( flOffsets, sizeof(flOffsets) );
		char nodetype = buf.GetChar();
		unsigned short nodeinfo = buf.GetUnsignedShort();
		short zone = buf.GetShort();

		//	gLua->Msg("%d/%d (%d) %d %d - %f %f %f\n", i, numNodes, gLua->Top(), table->isTable(), node->isTable(), vector.x, vector.y, vector.z);

		gLua->PushReference(vectormetaref);
		gLua->Push(vector.x); gLua->Push(vector.y); gLua->Push(vector.z);
		gLua->Call(3, 1);
		
		ILuaObject *vec = gLua->GetReturn(0);
		//	gLua->Msg("vec: %d node: %d\n", vec->isUserData(), node->isTable());
		node->SetMember("pos", vec);
		vec->UnReference();

		node->SetMember("yaw", yaw);

		ILuaObject *offsets = gLua->GetNewTable();
		for(int y=0; y < NUM_HULLS; y++)
			offsets->SetMember((float)y+1,flOffsets[y]);

		node->SetMember("offset", offsets);
		offsets->UnReference();

		node->SetMember("type", (float)nodetype);
		node->SetMember("info", (float)nodeinfo);		
		node->SetMember("zone", (float)zone);

		ILuaObject *nodeneighbors = gLua->GetNewTable();
		node->SetMember("neighbor", nodeneighbors);
		node->SetMember("numneighbors", (float)0);
		nodeneighbors->UnReference();

		ILuaObject *nodelinks = gLua->GetNewTable();
		node->SetMember("link", nodelinks);
		node->SetMember("numlinks", (float)0);
		nodelinks->UnReference();

		// re-grab the table from the stack, to make sure we don't lose the pointer after a while
		table = gLua->GetObject();
		//		gLua->Msg("top (%d) %d %d\n", gLua->Top(), table->isTable(), node->isTable());
		table->SetMember((float)i+1, node);
		table->UnReference();
		node->UnReference();
	}

	int numLinks = buf.GetInt();

	//gLua->Msg("%d links (%d nodes)\n", numLinks, numNodes);

	table = gLua->GetNewTable();
	table->Push();
	table->UnReference();

	for(int i=0; i < numLinks; i++)
	{

		ILuaObject *link = gLua->GetNewTable();

		int srcID = buf.GetShort();
		int destID = buf.GetShort();

		ILuaObject *nodes = gLua->GetObject(-2);
		ILuaObject *nodesrc = nodes->GetMember((float)srcID+1);
		ILuaObject *nodedest = nodes->GetMember((float)destID+1);
		if(nodesrc && nodedest)
		{
			ILuaObject *nodelinks = nodesrc->GetMember("link");
			ILuaObject *nodeneighbors = nodesrc->GetMember("neighbor");
			int numlinks = nodesrc->GetMemberInt("numlinks");
			int numneighbors = nodesrc->GetMemberInt("numneighbors");

			nodeneighbors->SetMember((float)++numneighbors, nodedest);
			nodesrc->SetMember("numneighbors", (float)numneighbors);
			nodelinks->SetMember((float)++numlinks, link);
			nodesrc->SetMember("numlinks", (float)numlinks);
			link->SetMember("src", nodesrc);

			nodelinks->UnReference();
			nodesrc->UnReference();

			nodelinks = nodedest->GetMember("link");
			nodeneighbors = nodedest->GetMember("neighbor");
			numlinks = nodedest->GetMemberInt("numlinks");
			numneighbors = nodedest->GetMemberInt("numneighbors");

			nodeneighbors->SetMember((float)++numneighbors, nodesrc);
			nodedest->SetMember("numneighbors", (float)numneighbors);
			nodelinks->SetMember((float)++numlinks, link);
			nodedest->SetMember("numlinks", (float)numlinks);
			link->SetMember("dest", nodedest);

			nodelinks->UnReference();
			nodedest->UnReference();
		} else {
			gLua->Msg("Unknown link source or destination %d %d\n", srcID, destID);
		}
		nodes->UnReference();

		byte acceptedMoves[NUM_HULLS];
		buf.Get( acceptedMoves, sizeof(acceptedMoves) );

		ILuaObject *moves = gLua->GetNewTable();
		for(int y=0; y < NUM_HULLS; y++)
			moves->SetMember((float)y+1,(float)acceptedMoves[y]);

		link->SetMember("move", moves);
		moves->UnReference();

		table = gLua->GetObject();
		table->SetMember((float)i+1, link);
		table->UnReference();
		link->UnReference();
	}

	gLua->FreeReference(vectormetaref);
	return 2;
}

LUA_FUNCTION(ParseGraphFile)
{
	ILuaInterface *gLua = Lua();

	gLua->CheckType(1, GLua::TYPE_STRING);

	const char *file = gLua->GetString(1);

	HANDLE hFile = CreateFileA(file, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if (hFile == INVALID_HANDLE_VALUE) 
	{ 
		gLua->Msg("Unable to open file %s\n", file);
		return 0;
	}

	DWORD sz = GetFileSize(hFile, NULL);
	char *buffer = new char[sz+1];

	DWORD dwBytesRead;
	ReadFile(hFile, buffer, sz, &dwBytesRead, NULL);

	CloseHandle(hFile);

	CUtlBuffer buf;
	buf.SetExternalBuffer(buffer, sz+1, dwBytesRead);

	int created = ParseGraph(buf, gLua);

	buf.Clear();
	delete buffer;

	return created;
}

int Start(lua_State *L)
{
	ILuaInterface *gLua = Lua();

	gLua->SetGlobal("ParseGraphFile", ParseGraphFile);

	return 0;
}

int Close(lua_State *L)
{
	return 0;
}